/*
 * rarpd - monitor/respond to RARP requests
 */
#include <stdio.h>
#include <stdlib.h>
#ifdef HAVE_STRING_H
#include <string.h>
#else
#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif
#endif
#include <unistd.h>
#include <sys/types.h>
#include <sys/param.h>
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#include <signal.h>

#include <netinet/in.h>
#ifdef HAVE_NET_IF_ARP_H
/* For SIOCSARP */
#include <net/if_arp.h>
#endif
#include <pcap.h>
#include <libnet.h>
#ifdef HAVE_CTYPE_H
#include <ctype.h>
#endif
#include <netdb.h>
#ifdef HAVE_PCAP_NAMEDB_H
#include <pcap-namedb.h>
#endif
#include <sys/ioctl.h>
#ifdef HAVE_SYS_SOCKIO_H
#include <sys/sockio.h>
#endif
#include <errno.h>
#ifndef MAXHOSTNAMELEN
#define MAXHOSTNAMELEN 256
#endif

#ifdef HAVE_SIGSET
#define SIGNALFUNC sigset
#else
#define SIGNALFUNC signal
#endif

/* Forwards */

#define HASHNAMESIZE 4096

struct enamemem {
	u_short e_addr0;
	u_short e_addr1;
	u_short e_addr2;
	char *e_name;
	u_char *e_nsap;			/* used only for nsaptable[] */
	struct enamemem *e_nxt;
};

struct enamemem enametable[HASHNAMESIZE];

int vflag;			/* be verbose */

/* fwd */
char *etheraddr_string(register const u_char *ep);
void init_addrtoname(bpf_u_int32 localnet, bpf_u_int32 mask);
char *intoa(bpf_u_int32 addr);

char *program_name;

extern int ether_ntohost (char *hostname, struct ether_addr *e);

/* Forwards */
extern void usage(void);

#define NTOHL(x)    (x) = ntohl(x)
#define NTOHS(x)    (x) = ntohs(x)
#define EXTRACT_16BITS(p) \
        ((u_short)ntohs(*(u_short *)(p)))
#define SHA(ap) ((ap)->ar_sha)
#define THA(ap) ((ap)->ar_tha)
#define SPA(ap) ((ap)->ar_spa)
#define TPA(ap) ((ap)->ar_tpa)

#define  ETHERADDRLEN (ETHER_ADDR_LEN)
#define  IPADDRLEN  (4)
static pcap_t *pd;
static struct link_int *ln;
static char *device;
static unsigned int myip;
static u_char myhwaddr[ETHERADDRLEN];

extern int optind;
extern int opterr;
extern char *optarg;

#define UPCASE 0
#define DOWNCASE 1
char *fixcase(char *d, char *s, int n, int dcase)
{
  char *o = d;
  char c;

  if ((s == 0) || (d == 0)) return 0; /* just to be safe */
  while (n) {
    c = *s++;
    if (dcase && isupper(c)) c = tolower(c);
    if (!dcase && islower(c)) c = toupper(c);
    *d++ = c;
    if (c == 0) break;              /* quit after transferring null */
    if (--n == 0) *(d-1) = 0;       /* make sure null terminated */
  }
  return o;
}

void set_arp(unsigned int herip, u_char *herhwaddr) 
{
  int fd;
  struct arpreq ar;
  struct sockaddr_in *si;

  si = (struct sockaddr_in *)&ar.arp_pa;
#ifdef linux
  strncpy(ar.arp_dev,device,15);
#endif
  memcpy(ar.arp_ha.sa_data, (char *)herhwaddr, ETHERADDRLEN);
  memcpy((char *) &si->sin_addr, &herip, IPADDRLEN);

  ar.arp_flags = 0;

  if ((fd = socket(AF_INET, SOCK_DGRAM, 0)) == -1)
      return;

  ioctl(fd, SIOCDARP, (caddr_t)&ar); 

  ioctl(fd, SIOCSARP, (caddr_t)&ar);

  close(fd);
  return;
}

void send_arpreply(unsigned int herip, u_char *herhwaddr)
{
    char ebuf[PCAP_ERRBUF_SIZE];
    u_char buf[ARP_H+ETH_H];

    memset(buf, 0, ARP_H + ETH_H);

    /*
     *  Ethernet header
     */
    build_ethernet(herhwaddr, myhwaddr, ETHERTYPE_ARP, NULL, 0, buf);

    /*
     *  ARP header
     */
    build_arp(ARPHRD_ETHER,
        ETHERTYPE_IP,
        ETHERADDRLEN,
        IPADDRLEN,
        ARPOP_REPLY,
        myhwaddr,
	(u_char *)&myip,
	herhwaddr,
	(u_char *)&herip,
        NULL,
        0,
        buf + ETH_H);

    write_link_layer(ln, (const u_char *)device, buf, ARP_H + ETH_H);
}

void send_rarpreply(unsigned int herip, u_char *herhwaddr)
{
    char ebuf[PCAP_ERRBUF_SIZE];
    u_char buf[ARP_H+ETH_H];

    memset(buf, 0, ARP_H + ETH_H);

    /*
     *  Ethernet header
     */
    build_ethernet(herhwaddr, myhwaddr, ETHERTYPE_REVARP, NULL, 0, buf);

    /*
     *  ARP header
     */
    build_arp(ARPHRD_ETHER,
        ETHERTYPE_IP,
        ETHERADDRLEN,
        IPADDRLEN,
        ARPOP_REVREPLY,
        myhwaddr,
	(u_char *)&myip,
	herhwaddr,
	(u_char *)&herip,
        NULL,
        0,
        buf + ETH_H);

    write_link_layer(ln, (const u_char *)device, buf, ARP_H + ETH_H);
}

void
arp_print(u_char *user, const struct pcap_pkthdr *h, const u_char *bp)
{
  const u_char *snapend, *packetp;
  u_int caplen=h->caplen;
  register const struct arpheader *ap;
  register u_short pro, hrd, op;
  register struct arpheader *eap;
  unsigned int herip;
  struct hostent *hp;
  static u_char herhwaddr[ETHERADDRLEN];

  packetp = bp;
  snapend = bp + caplen;
  caplen -= sizeof(struct ethheader);
  bp += sizeof(struct ethheader);

  ap = (struct arpheader *) bp;

  pro = EXTRACT_16BITS (&ap->ar_pro);
  hrd = EXTRACT_16BITS (&ap->ar_hrd);
  op = EXTRACT_16BITS (&ap->ar_op);

  eap = (struct arpheader *) ap;

  if (ap->ar_hln == sizeof (SHA (eap)) &&
      ap->ar_pln == sizeof (SPA (eap)))
    {
      char toret[1024];
	    
      switch(op) {
      case ARPOP_REVREQUEST:
	    if ((!ether_ntohost(toret, (struct ether_addr *)THA (eap))) &&
		(hp = gethostbyname(toret)) && (hp->h_addrtype == AF_INET) &&
		(hp->h_length == IPADDRLEN)) 
	      {
		struct in_addr in;
		char **p;
		
		in.s_addr = herip;
		p = hp->h_addr_list;

		/* XXX fix to find correct ip address if we need to from list */
		memcpy((char *) &herip, *p, IPADDRLEN);
		memcpy(herhwaddr, SHA (eap), ETHERADDRLEN);
		
		if (vflag) 
		  (void)fprintf(stderr, "rarp who-is %s tell %s\n",
				etheraddr_string (THA (eap)),
				etheraddr_string (SHA (eap)));
		set_arp(herip, (u_char *)&herhwaddr);
		send_rarpreply(herip, (u_char *)&herhwaddr); 
	  }
	break;

	case ARPOP_REQUEST:
	if (memcmp((const u_char *)TPA (eap), (const u_char *)&myip, IPADDRLEN) == 0)
	  {
	    char hname[MAXHOSTNAMELEN];
	    memcpy((char *) &herip, TPA (eap), IPADDRLEN);
	    hp = gethostbyaddr((char *) &herip, 4, AF_INET);
	    if (hp)
	      fixcase(hname, (char *)hp->h_name, MAXHOSTNAMELEN, UPCASE);
	    if ((!ether_hostton(hname, &herhwaddr)) || 
		(fixcase(hname, hname, MAXHOSTNAMELEN, DOWNCASE) &&
		(!ether_hostton(hname, &herhwaddr)))) {
#if 0
	      if (vflag) 
		(void)fprintf(stderr, "arp who-has %s tell %s\n",
	                      intoa((const u_char *)TPA (eap)),
			      intoa((const u_char *)SPA (eap)));
#endif
	      send_arpreply(herip, (u_char *)&herhwaddr);
	    }
	  }
	break;
      }
    }
}

int
main(int argc, char **argv)
{
	register int op, i;
	char ebuf[PCAP_ERRBUF_SIZE];
	bpf_u_int32 localnet, netmask;
	register char *cp, *cmdbuf;
	struct bpf_program fcode;
	u_char *pcap_userdata;

	device = NULL;
	if ((cp = strrchr(argv[0], '/')) != NULL)
		program_name = cp + 1;
	else
		program_name = argv[0];

	opterr = 0;
	while (
	    (op = getopt(argc, argv, "i:v")) != EOF)
		switch (op) {

		case 'v':
			++vflag;
			break;

		case 'i':
			device = optarg;
			break;

		default:
			usage();
			/* NOTREACHED */
		}

	if (!vflag) {
	  int pid;
	  pid = fork();
	  if (pid < 0) 
	    exit(1);
	  else if (pid != 0) 
	    exit(0);
	}

	if (device == NULL) {
	  device = pcap_lookupdev(ebuf);
	  if (device == NULL) {
	    fprintf(stderr, "%s", ebuf);
	    exit(255);
	  }
	}

	ln = open_link_interface(device, ebuf);
	if (ln == NULL) {
	  fprintf(stderr, "%s\n", ebuf);
	  exit(255);
	}
	pd = pcap_open_live(device, 68, 0, 1000, ebuf);
	if (pd == NULL) {
	  fprintf(stderr, "%s\n", ebuf);
	  exit(255);
	}
	i = pcap_snapshot(pd);
	if (pcap_lookupnet(device, &localnet, &netmask, ebuf) < 0) {
	  localnet = 0;
	  netmask = 0;
	}
	myip = get_ipaddr(ln, device, ebuf);
	memcpy(myhwaddr, (char *)get_hwaddr(ln, device, ebuf), ETHERADDRLEN);

	/*
	 * Let user own process after socket has been opened.
	 */
	setuid(getuid());
	cmdbuf = strdup("rarp or arp");

	if (pcap_compile(pd, &fcode, cmdbuf, 1, netmask) < 0) {
	  fprintf(stderr, "%s", pcap_geterr(pd));
	  exit(255);
	}
	init_addrtoname(localnet, netmask);

	if (pcap_setfilter(pd, &fcode) < 0) {
	  fprintf(stderr, "%s", pcap_geterr(pd));
	  exit(255);
	}
	pcap_userdata = 0;
	if (vflag)
	  (void)fprintf(stderr, "%s: listening on %s\n",
			program_name, device);
	(void)fflush(stderr);
	if (pcap_loop(pd, -1, arp_print, pcap_userdata) < 0) {
		(void)fprintf(stderr, "%s: pcap_loop: %s\n",
		    program_name, pcap_geterr(pd));
		exit(1);
	}
	pcap_close(pd);
	exit(0);
}

void
usage(void)
{
	extern char pcap_version[];

	(void)fprintf(stderr, "%s\n", program_name);
	(void)fprintf(stderr, "libpcap version %s\n", pcap_version);
	(void)fprintf(stderr,
"Usage: %s [-fv] [ -i interface ]\n", program_name);
	exit(-1);
}

static bpf_u_int32 netmask;

static char hex[] = "0123456789abcdef";


/* Find the hash node that corresponds the ether address 'ep' */

static struct enamemem *
lookup_emem(const u_char *ep)
{
	register u_int i, j, k;
	struct enamemem *tp;

	k = (ep[0] << 8) | ep[1];
	j = (ep[2] << 8) | ep[3];
	i = (ep[4] << 8) | ep[5];

	tp = &enametable[(i ^ j) & (HASHNAMESIZE-1)];
	while (tp->e_nxt)
		if (tp->e_addr0 == i &&
		    tp->e_addr1 == j &&
		    tp->e_addr2 == k)
			return tp;
		else
			tp = tp->e_nxt;
	tp->e_addr0 = i;
	tp->e_addr1 = j;
	tp->e_addr2 = k;
	tp->e_nxt = (struct enamemem *)calloc(1, sizeof(*tp));
	if (tp->e_nxt == NULL) {
	  fprintf(stderr, "lookup_emem: calloc");
	  exit(255);
	}

	return tp;
}

char *
etheraddr_string(register const u_char *ep)
{
	register u_int i, j;
	register char *cp;
	register struct enamemem *tp;
	char buf[sizeof("00:00:00:00:00:00")];

	tp = lookup_emem(ep);
	cp = buf;
	if ((j = *ep >> 4) != 0)
		*cp++ = hex[j];
	*cp++ = hex[*ep++ & 0xf];
	for (i = 5; (int)--i >= 0;) {
		*cp++ = ':';
		if ((j = *ep >> 4) != 0)
			*cp++ = hex[j];
		*cp++ = hex[*ep++ & 0xf];
	}
	*cp = '\0';
	tp->e_name = strdup(buf);
	return (tp->e_name);
}

static struct etherlist {
	u_char addr[ETHERADDRLEN];
	char *name;
} etherlist[] = {
	{{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff }, "Broadcast" },
	{{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, NULL }
};

/*
 * Initialize the ethers hash table.  We take two different approaches
 * depending on whether or not the system provides the ethers name
 * service.  If it does, we just wire in a few names at startup,
 * and etheraddr_string() fills in the table on demand.  If it doesn't,
 * then we suck in the entire /etc/ethers file at startup.  The idea
 * is that parsing the local file will be fast, but spinning through
 * all the ethers entries via NIS & next_etherent might be very slow.
 */
static void
init_etherarray(void)
{
	register struct etherlist *el;
	register struct enamemem *tp;
	char name[256];

	/* Hardwire some ethernet names */
	for (el = etherlist; el->name != NULL; ++el) {
		tp = lookup_emem(el->addr);
		/* Don't override existing name */
		if (tp->e_name != NULL)
			continue;

                /* Use yp/nis version of name if available */
                if (ether_ntohost(name, (struct ether_addr *)el->addr) == 0) {
                        tp->e_name = strdup(name);
			continue;
		}

		tp->e_name = el->name;
	}
}

void
init_addrtoname(bpf_u_int32 localnet, bpf_u_int32 mask)
{
	netmask = mask;
	init_etherarray();
}

char *
intoa(bpf_u_int32 addr)
{
        register char *cp;
        register u_int byte;
        register int n;
        static char buf[sizeof(".xxx.xxx.xxx.xxx")];

        NTOHL(addr);
        cp = &buf[sizeof buf];
        *--cp = '\0';

        n = 4;
        do {
                byte = addr & 0xff;
                *--cp = byte % 10 + '0';
                byte /= 10;
                if (byte > 0) {
                        *--cp = byte % 10 + '0';
                        byte /= 10;
                        if (byte > 0)
                                *--cp = byte + '0';
                }
                *--cp = '.';
                addr >>= 8;
        } while (--n > 0);

        return cp + 1;
}
